require "ncurses"
require 'thread'

require 'imap/emaiimap'
require 'accounts/accounts'
require 'email/emaiemail'
require 'frontend/emaiwindows'


class EmaiFrontend
  include Ncurses
  include EmaiEmail
  attr_accessor :cols, :lines, :accounts, :colors, :display_mutex
  def initialize()
    @cols  = 0
    @lines = 0
    @accounts      = {}
    @ruby_threads  = []
    @display_mutex = Mutex.new
  end
  
  def init_curses() # initialize ncurses
    Ncurses.initscr
    Ncurses.cbreak                   # provide unbuffered input
    Ncurses.noecho                   # turn off input echoing
    Ncurses.nonl                     # turn off newline translation
    Ncurses.stdscr.intrflush(false)  # turn off flush-on-interrupt
    Ncurses.stdscr.scrollok(true)
    Ncurses.stdscr.keypad(true)      # turn on keypad mode
    Ncurses::nodelay(Ncurses::stdscr, TRUE)
    Ncurses::curs_set(0)

    if (Ncurses::has_colors?())
      # XXX How to use 16 or more colors
      Ncurses::use_default_colors()
      Ncurses::start_color()
      my_bg =  -1 # Default Background of ncurses is -1
      Ncurses::init_pair(1, Ncurses::COLOR_BLACK, Ncurses::COLOR_CYAN)
      Ncurses::init_pair(2, Ncurses::COLOR_CYAN, my_bg)
      Ncurses::init_pair(3, my_bg,Ncurses::COLOR_GREEN)
      Ncurses::init_pair(4, Ncurses::COLOR_GREEN, my_bg)
      Ncurses::init_pair(5, my_bg, Ncurses::COLOR_MAGENTA)
      Ncurses::init_pair(6, Ncurses::COLOR_MAGENTA, my_bg)
      Ncurses::init_pair(7, my_bg, Ncurses::COLOR_WHITE)
      Ncurses::init_pair(8, Ncurses::COLOR_WHITE, my_bg)
      Ncurses::init_pair(9, my_bg, Ncurses::COLOR_BLUE)
      Ncurses::init_pair(10, Ncurses::COLOR_BLUE, my_bg)
      Ncurses::init_pair(11, my_bg, Ncurses::COLOR_YELLOW)
      Ncurses::init_pair(12, Ncurses::COLOR_YELLOW, my_bg)
      Ncurses::init_pair(13, my_bg, Ncurses::COLOR_RED)
      Ncurses::init_pair(14, Ncurses::COLOR_RED, my_bg)
    end
   
    @cols  = Ncurses.COLS
    @lines = Ncurses.LINES
  end


  def shutdown()
    Ncurses.echo
    Ncurses.nocbreak
    Ncurses.nl
    Ncurses.endwin
  end

  def error(args)
    args_s = ""
    args.class == "Array" ? args_s = args.join(" ") : args_s = args.to_s
    $logger.info "INFO: #{args_s}"
    @error_window.error(args_s)
  end

  def start()
    right_y    = @cols / 4
    right_cols = @cols - right_y
    head_cols  = 1
    foot_cols  = 1
    mlines     = @lines - foot_cols - head_cols

    items = []
    @accounts.each_value { |a| items << a}
    @account = items[0]

    # Layout
    #
    # -------------------
    # | head (unused)   |
    # |-----------------|
    # |l  |             |
    # |e  |             |
    # |f  | right       |
    # |t  |             |
    # |-----------------|
    # | foot            |
    # -------------------
    @head  = EmaiWindowMenu.new([head_cols, @cols, 0, 0], ["file", "folder", "help", "about"], 0, 1)
    @left  = EmaiWindowFolders.new([mlines, @cols - right_cols, head_cols , 0], items, 1, 0)
    @right = EmaiWindowEmails.new([mlines, right_cols, head_cols, right_y], [], 1, 0)
    @foot  = EmaiWindowStatus.new([foot_cols, @cols, @lines - foot_cols, 0])

    @windows = []; @menu_windows = []
    @error_window = @foot
    @windows << @left << @right << @head << @foot
    @menu_windows << @left << @right # << @head
    @windows.each {|w|
      w.error_window = @error_window
      w.w.intrflush(false)
      w.w.keypad(true)
      w.w.nodelay(true)
      w.w.keypad(true)

      # we do our own scrolling
      #w.w.scrollok(true)
    }

    @menu_windows.each { |w| w.rewin }
  end

  def looping
    window = 0
    @window_history = []
    @res_history    = []
    @ruby_threads << Thread.new do
      # update the Folderlist
      loop do
        sleep 300
        $logger.info "Updating folder list"
        @left.items.each { |i|
          if i.kind_of?(Folder) then
            i.count_msgs
          end
        }
        @display_mutex.synchronize do
          @left.h_item()
        end
        $logger.info "Updating folder list done"
      end
    end

    # main input loop calls window loops
    loop {

      begin
        w = @menu_windows[window]
        # pass to current window
        res = w.input_loop
        error(res.to_s)
        data = res[1] - 1 if(res[1] && res[1] != "")

        @error_window.info(res[0])
        case res[0]
        when "QUIT"
          break

        when "EDIT EMAIL"
          message = w.items[data]
          mail = message.header.decoded + "\n" + message.body_text
          edit(mail)
          @windows.each {|w| w.w.refresh}

        when "GROUP REPLY EMAIL"
          group_reply_mail(w.items[data], @account)
          @windows.each {|w| w.w.refresh}

        when "LIST REPLY EMAIL"
          list_reply_mail(w.items[data], @account)
          @windows.each {|w| w.w.refresh}

        when "FORWARD EMAIL"
          forward_email(w.items[data], @account)
          @windows.each {|w| w.w.refresh}

        when "REPLY EMAIL"
          reply_mail(w.items[data], @account)
          @account.mark(w.items[data], [:Answered])
          @windows.each {|w| w.w.refresh}

        when "WRITE EMAIL"
          write_email(@account)
          @windows.each {|w| w.w.refresh}

        when "SHELL"
          Ncurses.def_prog_mode            # save current tty modes
          Ncurses.endwin                   # restore original tty modes
          system("$SHELL")
          @windows.each {|w| w.w.refresh}

        when "NEXT WINDOW"
          if @menu_windows.length <= window + 1
            window = 0
          else
            window += 1
          end
          if window == 0
            @menu_windows[0].h_item()
          end
          @menu_windows[1].clean
          @menu_windows.each {|w| w.rewin }

        when "OPEN FOLDER"
          w = @menu_windows[window]
          folder = w.items[data]
          name   = folder.name
          if folder.kind_of? Account
            items = []
            @accounts.each_value { |a| items << a}
            w.items = items
            @account = @accounts[name]
            @account.open
            if @account.folders
              w.items = w.items[0 .. data] + @account.folders.collect{ |f| f } + w.items[data + 1 .. -1]
            else
              error("no Folders")
            end
            w.h_item(data + 2)
          else # an email folder
            if @account.select(folder)
              messages = @account.get_messages
              if messages
                threads = messages.threaded
                window = 1
                w = @menu_windows[window]
                w.items = threads
                w.folder = folder
                if threads.first_to_show
                  show = threads.first_to_show.threaded_index + 1
                else
                  show = threads.length
                end
                w.h_item(show)
              else
                @menu_windows[1].w.clear
              end
            end
          end

        when "CLOSE FOLDER"
          items = []
          @accounts.each_value { |a| items << a}
          w.items = items
          w.h_item(data)

        when "READ EMAIL"
          w.show_message

        when "MARK EMAIL DELETED"
          @account.mark(w.items[data], [:Deleted])
          w.next_item

        when "MARK EMAIL UNDELETED"
          @account.unmark(w.items[data], [:Deleted])
          w.next_item

        else
          error(*res)
        end
        @res_history << res
      rescue => e
        $logger.error e.inspect
        error(e.message)
      end
    }
  end
end

#vim: ts=2 sw=2 expandtab
