#
# Ncurses::Window based classes
#



class EmaiWindow

  include Ncurses
  attr_accessor :error_window
  attr_reader :w

  # XXX should be @@config
  @@sleep_time = 0.02
  @@encoding = "UTF-8"

  def initialize(*args)
    init_colors()
    @error_window = nil
    @args = *args
    @height = @args[0]
    @breadth = @args[1]
    @w = Ncurses::WINDOW.new(*@args)
    @border = [1]*8
    @border_lines = 2
    @w.wrefresh
  end

  def init_colors
    @colors = {
                          "none"     => 0,
      "cyan_bg"    => 1,  "cyan"     => 2,
      "green_bg"   => 3,  "green"    => 4,
      "magenta_bg" => 5,  "magenta"  => 6,
      "white_bg"   => 7,  "white"    => 8,
      "blue_bg"    => 9,  "blue"     => 10,
      "yellow_bg"  => 11, "yellow"   => 12,
      "red_bg"     => 13, "red"      => 14
    }
  end

  def rewin
    @w.border(*@border)
    @w.wrefresh
  end

  def mv_print_color(y, x, format, str, color="none")
    # get rid of control characters
    if @@encoding == "UTF-8"
      r = Regexp.new('[[:cntrl:]]', nil, 'U')
    else
      r = Regexp.new('[[:cntrl:]]')
    end
    str.gsub!(r,'?')
    
    if color == "reverse" || (not Ncurses::has_colors?())
      w.wattron(Ncurses::A_REVERSE)
      w.mvwprintw(y, x, format, str)
      w.wattroff(Ncurses::A_REVERSE)
    elsif Ncurses::has_colors?() && color != "none"
      color_n = @colors[color] || 0
      w.wattron(Ncurses::COLOR_PAIR(color_n));
      w.mvwprintw(y, x, format, str)
      w.wattroff(Ncurses::COLOR_PAIR(color_n));
    else
      w.mvwprintw(y, x, format, str)
    end
  end

  def readline
    data = ""
    while c = @w.getch do
      case c
      when -1
        sleep @@sleep_time
      when KEY_LEFT
        @error_window.w.move(-1)
      when KEY_ENTER, ?\n, ?\r
        return data
      when KEY_BACKSPACE, 263
        data.chop!
        @error_window.del_char(c.chr)
      else
        @error_window.add_char(c.chr)
        data << c.chr
      end
    end
  end
end


class EmaiWindowStatus < EmaiWindow
  # override some methods because we can't use borders (cols < 3)
  def initialize(args)
    init_colors
    @args = args
    @w = Ncurses::WINDOW.new(*@args)
    @height = @args[0]
    @breadth = @args[1]
    reset
    info("Hallo Dave <(o)>")
  end

  def rewin
    @w.wrefresh
  end

  def clear
    reset
    @w.clear
  end

  def reset
    @x = 1
    mv_print_color(0, 0, "%-#{@breadth}s", " ")
  end

  def add_char(char)
    y = 0;
    mv_print_color(y, @x, "%s", char)
    @x += 1;
    rewin
  end

  def add_string(string)
    y = 0;
    mv_print_color(y, @x, "%s", string)
    @x += string.length;
    rewin
  end

  def error(msg)
    reset
    y = 0; x = 2;
    mv_print_color(y, x, "ERROR: %s", msg)
    rewin
  end

  def info(msg)
    reset
    y = 0; x = 2;
    mv_print_color(y, x, "%s", msg)
    rewin
  end
end


class EmaiWindowMenu < EmaiWindow

  attr_accessor :items, :hitem, :read_email

  def initialize(args, items, y, x)
    super(*args)
    @args = args
    @items = items || []
    @hitem = 0
    @y = y ; @x = x
    @read_email = false
    @search_string = ""
    @visible = @height - @border_lines
  end

  def get_offset(items, visible, i)
    if items.length < visible / 2
      return 0
    elsif i >= (items.length - (visible / 2))
      return [items.length - visible, 0].max
    elsif i > visible / 2
      return i - visible / 2
    else
      return 0
    end
  end

  def read_line(new_line = ?\n)
    input = ""
    loop {
      c = Ncurses.wgetch(w)
      case c

      when KEY_ENTER, ?\n, ?\r
        break

      else
        if c < 256 && c > 0
          @error_window.add_char(c.chr)
          input += c.chr
        end
      end
    }
    input
  end

  def scroll(n)
    case n
    when "END"
      h_item(@items.length)
    when "BEGIN"
      h_item(0)
    else
      h_item(@hitem + n)
    end
  end

  # Subclasses will pass to this if they don't handle a key themselves.
  # Thus a common set of keys that works in all windows is handled here.
  def handle_key(c)
    case c

    # we call getch nonblocking to give other threads a chance to run
    when -1
      sleep @@sleep_time

    when KEY_UP, ?k
      scroll(-1)

    when KEY_DOWN, ?j
      scroll(1)

    when KEY_NPAGE
      scroll(@visible - 1)

    when KEY_PPAGE
      scroll(- @visible + 1)

    when ?G
      scroll("END")

    when ?H
      scroll("BEGIN")

    when 32  # SPACE
      scroll((@visible - 3) / 2)

    when ?m
      return ["WRITE EMAIL", @hitem]

    when 262 # Home
      h_item(0)

    when ?G, 360 # End
      h_item(@items.length)

    when ?c, ?\t
      return ["NEXT WINDOW", ""]

    when ?\q
      @error_window.info("Don't do that Dave")
      return ["QUIT", ""]

    when ?\:
      @error_window.reset
      @error_window.add_char(c.chr)
      data = readline
      # jump to line
      if data =~ /^(\d+)$/
        h_item(data.to_i)
      end

    else
      @error_window.info("NOTHING FOR KEY: #{c.chr}")
      $logger.warn "NOTHING FOR KEY: #{c.chr}"
    end
    
    return ["LOOP"]
  end

  def input_loop
    loop do
      data = handle_key w.getch
      rewin
      unless data[0] == "LOOP"
        return data # something the top loop got to handle
      end
    end
  end
end


class EmaiWindowFolders < EmaiWindowMenu
  def initialize(*args)
    super(*args)
    @border = [0]*8
    @offset = 0
    @hitem  = 0
    h_item(1)
    rewin
  end

  def h_item(i=@hitem)
    x = 1; y = 0; z = 0
    visible = (@height - y - 2)
    if i < 1 then i = 1 end
    if i > @items.length then i = @items.length end

    @offset = get_offset(items, visible, i)
    @hitem = i
    show_item = i - @offset
    @items[@offset .. @offset + visible].each { |item| z += 1 ; y += @y; x += @x
      case item
      when Account
        itemname = sprintf("%-#{@breadth - 2}s",item.name.slice(0, @breadth - 3))
      when Folder
        stats = "%11s" % item.stats_to_s
        itemname = sprintf("%#{stats.length}s %-#{@breadth - 3 - stats.length}s", stats, item.to_s)
      else # should not happen
        itemname = item.class
      end

      if z == show_item
        mv_print_color(y, x, "%s", itemname, "reverse")
      else
        mv_print_color(y, x, "%s", itemname, "none")
      end
      x += itemname.length if @y == 0
    } if @items[@offset .. @offset + visible]

    if @items.length < (@offset + visible)
      (@items.length + 1 - @offset).upto(visible) { |y|
        mv_print_color(y, x, "%-#{@breadth - 2}s", "~", "blue")
      }
    end
    rewin
  end

  def handle_key(c)
    case c

    when -1
      sleep @@sleep_time

    when KEY_RIGHT, KEY_ENTER, ?\n, ?\r, ?l
      return ["OPEN FOLDER", @hitem]

    when KEY_LEFT, ?h
      return ["CLOSE FOLDER", @hitem]

    # find
    when ?/
      @error_window.reset
      @error_window.add_char(c.chr)
      @search_string = read_line
      i = @hitem
      @items[i .. -1].each { |item|
        i += 1
        if item.to_s =~ /#{@search_string}/
          h_item(i)
          break
        end

      }

    # find next
    when ?n
      i = @hitem
      @items[i .. -1].each { |item|
        i += 1
        if item.to_s =~ /#{@search_string}/
          h_item(i)
          break
        end
      }

    # create folder
    when ?C
      item = @items[@hitem - 1]
      case item
      when Account
        account = item
        parent_folder = "INBOX."
      when Folder
        account = item.account
        parent_folder = item.name + "."
      else # should not happen
        raise Error
      end
      @error_window.add_string("Create folder: #{parent_folder}")
      new_folder = read_line
      account.create(parent_folder + new_folder)
      return ["OPEN FOLDER", @items.index(account) + 1]

    # delete folder
    when ?D, 330 # del
      item = @items[@hitem - 1]
      case item
      when Account
        next
      when Folder
        account = item.account
        folder = item.name
      else # should not happen
        raise Error
      end
      @error_window.add_string("Delete folder: #{folder} [y/n] ? ")
      answer = read_line
      if answer == "y" || answer == "Y"
        account.delete folder
        return ["OPEN FOLDER", @items.index(account) + 1]
      else
        @error_window.clear
      end

    else
      return super(c)
    end

    return ["LOOP"]

  end
end


class EmaiWindowEmails < EmaiWindowMenu
  attr_accessor :read_email, :height, :breadth, :context, :offset, :visible, :folder
  def initialize(*args)
    super(*args)
    @offset  = 0
    @border  = [0]*8
    @context = 6
    @folder  = nil
    @hitem_from = nil
    @read_email = false
    @visible = @height - @context - @border_lines
    rewin
  end

  def clean
    1.upto(@height) { |l|
      w.mvwprintw(l, 1, "%#{@breadth - 2}s", "")
    }
  end

  def next_item
    i = @hitem + 1
    if i < 1 then i = 1 end
    if i > @items.length then i = @items.length end
    @hitem = i
    if read_email
      show_message
    else
      h_item()
    end
  end

  def show_message
    @visible = @height - @context - @border_lines
    @message = @items[@hitem - 1]

    @body   = @message.body_formatted(@breadth - 2)
    @header = @message.header.decoded.gsub(/\t/,"        ").gsub(/(.{#{@breadth - 2}})/, "\\1\n").split(/\r?\n/)
    @header << ""
    @read_email_offset = @header.length
    @read_email = true
    @search_string = ""
    $logger.debug @header if $debug > 0
    display
  end

  def scroll(n)
    if read_email
      case n
      when "END"
        scroll_message(@body.length + @header.length - @visible - @read_email_offset + 1)
      when "BEGIN"
        scroll_message(-@read_email_offset)
      else
        scroll_message(n)
      end
    else
      super n
    end
  end

  def scroll_message(n)
    @read_email_offset = [@body.length + @header.length, [0, @read_email_offset +  n].max].min
    display
  end

  def display
    x = 1
    y = @context + 1
    height  = @height
    visible = @visible
    format = "%-#{@breadth - 2}s"

    h = @header[@read_email_offset .. @read_email_offset + visible]
    message_offset = [@read_email_offset - @header.length, 0].max
    m = @body[message_offset .. message_offset + visible]
    
    h_item(@hitem, true, @context)

    mv_print_color(y, x, format, "-" * (@breadth - 2), "none")

    # header
    h.each { |l| y += 1
      mv_print_color(y, x, format, l, "green")
    } if h

    #y += 1
    #mv_print_color(y, x, format, "-" * (@breadth - 2), "none")
    
    in_signature = false
    # XXX @config which characters are quote characters
    # maybe intelligent version that figures it out itself ?
    quote_char   = ">"
    quote_regexp = "[>:]"
    colors = ["green","magenta", "cyan", "yellow", "blue", "cyan_bg"]
    
    # body
    m.each { |line| y += 1
      if line && line =~ /^$/
        in_signature = false
      elsif line && line =~ /^-- $/
        in_signature = true
      end

      if in_signature
        mv_print_color(y, x, format, line, "yellow")
      elsif match = /^(#{quote_regexp}\s*)+/.match(line)
        quotes = match[0]
        tail   = line[quotes.length, line.length]
        levels = quotes.split(/(?=#{quote_regexp})/)
        color  = "none"
        color  = colors[levels.length.modulo(colors.length) - 1]

        mv_print_color(y, x, format, tail, color)
        
        # XXX @config option to show quote characters
        #x_off = x
        #levels.each_index { |i|
        #  color = colors[i.modulo(colors.length)]
        #  mv_print_color(y, x_off, format, "#{quote_char} ", color)
        #  x_off += "#{quote_char} ".length
        #}
        #mv_print_color(y, x_off, format, tail, color)
      else
        mv_print_color(y, x, format, line, "none")
      end
    } if m

    # paint rest of screen with "~"
    o = y + 1
    o.upto(visible + @context + 1) { |y|
      mv_print_color(y, x, format, "~", "blue")
    } if @header.length + @body.length < (@read_email_offset + visible)

    rewin
  end

  def h_item(i=@hitem, read_email=false, visible=-1)
  ## higlight the current active element
    x = 1; y = 0; z = 0

    unless read_email
      @read_email_offset = 0
    end
    @read_email = read_email

    if visible < 0 then @visible = visible = (@height - y - 2) end
    if i < 1 then i = 1 end
    if i > @items.length then i = @items.length end
    @offset = get_offset(items, visible, i)
    
    @hitem       = i
    show_item    = i
    num_size     = 5
    flags_size   = 4
    date_size    = 16
    message_size = 4
    from_size    = 12
    subject_size = @breadth - flags_size - date_size - message_size - num_size - from_size - 7
    subject_size = 0 if subject_size < 0 
    z = @offset
    
    @items[@offset .. @offset + visible].each { |message| z += 1 ; y += @y; x += @x
      date = "missing"
      date = message.envelope.date if message.envelope.date
      from = "(empty)"
      from = message.envelope.from[0].mailbox if message.envelope.from && message.envelope.from[0].mailbox
      subject = message.threaded_subject || ""
      if ( subject == "" ||           # Top of a threads
           subject[0] == "-"[0] ||    # no Top available
           subject[0] == "+"[0] ||    # no Top available
           z == show_item ||          # highlited item
           z == @offset + 1 ||        # 1st visible item
           z == @offset + visible )   # last visible item
        subject += message.decoded_subject 
      end
      decoded = "%#{num_size}s %#{flags_size}s %#{from_size}s %-#{subject_size}s %-#{date_size}s %#{message_size}s" %
                 [z.to_s.slice(0, num_size),
                  message.flags.join("").slice(0, flags_size),
                  from.slice(0, from_size),
                  subject.slice(0, subject_size),
                  date.slice(0, date_size),
                  message.size_short.to_s.slice(0, message_size)
                 ]
      # colorize if either active or in visual mark range
      if z == show_item || (@hitem_from && (z.between?(show_item, @hitem_from) || z.between?(@hitem_from, show_item)))
        mv_print_color(y, x, "%s", decoded, "cyan_bg")
        name = ""
        name = message.envelope.from[0].name.unpack("M*") if message.envelope.from && message.envelope.from[0].name
        host = ""
        host = message.envelope.from[0].host if message.envelope.from && message.envelope.from[0].host
        @error_window.info("%s <%s@%s>: %s @ %s" %
          [name, from, host, message.decoded_subject, date])
      else
        mv_print_color(y, x, "%s", decoded, "none")
      end

      x += decoded.length if @y == 0
    }
    if @items.length < (@offset + visible)
      (@items.length + 1 - @offset).upto(visible) { |y|
        mv_print_color(y, x, "%-#{@breadth - 2}s", "~", "blue")
      }
    end
        
    rewin
  end

  def mark_deleted(from, to=nil)
    to ||= from
    @folder.account.mark(items[from .. to], [:Deleted])
  end

  def handle_key(c)
    case c

    when -1
       sleep @@sleep_time

    when ?\$
      $logger.info "EXPUNGE"
      @error_window.info("EXPUNGE")
      begin
        @folder.expunge
        @folder.count_msgs
        messages = @folder.get_messages
        if messages
          @items = messages.threaded
          h_item()
        else
          return ["NEXT WINDOW",""]
        end

      ensure
        $logger.info "EXPUNGE done: #{$!}"
      end

    # command mode
    when ?:
      @error_window.reset
      @error_window.add_char(c.chr)
      data = readline
      # jump to number
      if data =~ /^(\d+)$/
        h_item(data.to_i)
        # delete range
      elsif data =~ /^d(\d+)-(\d+)$/
        hitem = 1
        from = $~[1].to_i - 1
        to   = $~[2].to_i - 1
        mark_deleted(from, to)
        h_item(to)
      end

    when ?/
      @error_window.reset
      @error_window.add_char(c.chr)
      # read till ENTER
      @search_string = read_line
      i = @hitem
      @items[i .. -1].each { |message|
        i += 1
        if message.decoded_subject =~ /#{@search_string}/
          h_item(i)
          break
        end
      }

    # Next new message
    when ?\t
      i = 0 
      @items.each { |message|
        i += 1
        if message.flags.index("N")
          h_item(i)
          break
        end
      }

    when ?r
      return ["REPLY EMAIL", @hitem]

    when ?f
      return ["FORWARD EMAIL", @hitem]

    when ?g
      return ["GROUP REPLY EMAIL", @hitem]

    when ?L
      return ["LIST REPLY EMAIL", @hitem]

    when ?e
      return ["EDIT EMAIL", @hitem]

    when ?m
      return ["WRITE EMAIL", @hitem]

    when ?n
      i = @hitem
      @items[i .. -1].each { |message|
        i += 1
        if message.decoded_subject =~ /#{@search_string}/
          h_item(i)
          break
        end
      }

    when ?N
      i_item = [@hitem - 2, 0].max
      i_item.downto(0) { |i|
        message = @items[i]
        if message.decoded_subject =~ /#{@search_string}/
          h_item(i + 1)
          break
        end
      }

    when ?d, 330 # del
      if @hitem_from
        mark_deleted(@hitem_from - 1, @hitem - 1) 
        @hitem_from = nil
      else
        mark_deleted(@hitem - 1)
        # return ["MARK EMAIL DELETED", @hitem]
      end
      next_item()

    when ?v, ?V
      @hitem_from = @hitem

    when ?u
      return ["MARK EMAIL UNDELETED", @hitem]

    when KEY_LEFT, ?h
      if @read_email
        h_item()
      else

        return ["NEXT WINDOW",""]
      end

    when KEY_RIGHT, KEY_ENTER, ?\n, ?\r, ?l
      if @read_email
        @hitem = [@hitem + 1, @items.length].min
        return ["READ EMAIL", @hitem]
      else
        return ["READ EMAIL", @hitem]
      end

    else
      return super(c)
    end
    return ["LOOP"]
  end
end

# vim: ts=2 sw=2 expandtab
